@@ -14,6 +14,7 @@ import math
 import sys
 import warnings
 from collections.abc import Iterable
+from contextlib import AbstractContextManager, contextmanager
 from functools import cached_property
 from random import Random
 from sys import float_info
@@ -21,11 +22,12 @@ from types import ModuleType
 from typing import (
     TYPE_CHECKING,
     Any,
+    ClassVar,
     Literal,
     Optional,
+    TypeAlias,
     TypedDict,
     TypeVar,
-    Union,
 )
 
 from sortedcontainers import SortedSet
@@ -62,37 +64,92 @@ from hypothesis.internal.floats import (
     next_up,
 )
 from hypothesis.internal.intervalsets import IntervalSet
+from hypothesis.internal.observability import InfoObservationType, TestCaseObservation
 
 if TYPE_CHECKING:
-    from typing import TypeAlias
-
     from hypothesis.internal.conjecture.data import ConjectureData
     from hypothesis.internal.constants_ast import ConstantT
 
 T = TypeVar("T")
-_Lifetime: "TypeAlias" = Literal["test_case", "test_function"]
+LifetimeT: TypeAlias = Literal["test_case", "test_function"]
 COLLECTION_DEFAULT_MAX_SIZE = 10**10  # "arbitrarily large"
 
 
-# The available `PrimitiveProvider`s, and therefore also the available backends
-# for use by @settings(backend=...). The key is the name to be used in the backend=
-# value, and the value is the importable path to a subclass of PrimitiveProvider.
-#
-# See also
-# https://hypothesis.readthedocs.io/en/latest/strategies.html#alternative-backends-for-hypothesis.
-#
-# NOTE: the PrimitiveProvider interface is not yet stable. We may continue to
-# make breaking changes to it. (but if you want to experiment and don't mind
-# breakage, here you go!)
-AVAILABLE_PROVIDERS = {
+#: Registered Hypothesis backends. This is a dictionary where keys are the name
+#: to be used in |settings.backend|. The value of a key can be either:
+#:
+#: * A string corresponding to an importable absolute path of a
+#:   |PrimitiveProvider| subclass
+#: * A |PrimitiveProvider| subclass (the class itself, not an instance of the
+#:   class)
+#:
+#: Hypothesis will instantiate the corresponding |PrimitiveProvider| subclass
+#: when the backend is requested by a test's |settings.backend| value.
+#:
+#: For example, the default Hypothesis backend is registered as:
+#:
+#: .. code-block:: python
+#:
+#:    from hypothesis.internal.conjecture.providers import AVAILABLE_PROVIDERS
+#:
+#:    AVAILABLE_PROVIDERS["hypothesis"] = "hypothesis.internal.conjecture.providers.HypothesisProvider"
+#:    # or
+#:    AVAILABLE_PROVIDERS["hypothesis"] = HypothesisProvider
+#:
+#: And can be used with:
+#:
+#: .. code-block:: python
+#:
+#:     from hypothesis import given, settings, strategies as st
+#:
+#:     @given(st.integers())
+#:     @settings(backend="hypothesis")
+#:     def f(n):
+#:         pass
+#:
+#: Though, as ``backend="hypothesis"`` is the default setting, the above would
+#: typically not have any effect.
+#:
+#: For third-party backend authors, we strongly encourage ensuring that
+#: ``import hypothesis`` does not automatically import the expensive parts of
+#: your package, by:
+#:
+#: - setting a string path here, instead of a provider class
+#: - ensuring the registered hypothesis plugin path references a path which just
+#:   sets AVAILABLE_PROVIDERS and does not import your package
+AVAILABLE_PROVIDERS: dict[str, str | type["PrimitiveProvider"]] = {
     "hypothesis": "hypothesis.internal.conjecture.providers.HypothesisProvider",
     "hypothesis-urandom": "hypothesis.internal.conjecture.providers.URandomProvider",
 }
 # cache the choice_permitted constants for a particular set of constraints.
-CacheKeyT: "TypeAlias" = tuple[ChoiceTypeT, tuple[Any, ...]]
-CacheValueT: "TypeAlias" = tuple[tuple["ConstantT", ...], tuple["ConstantT", ...]]
+CacheKeyT: TypeAlias = tuple[ChoiceTypeT, tuple[Any, ...]]
+CacheValueT: TypeAlias = tuple[tuple["ConstantT", ...], tuple["ConstantT", ...]]
 CONSTANTS_CACHE: LRUCache[CacheKeyT, CacheValueT] = LRUCache(1024)
 
+
+_constants_integers = (
+    # powers of 2
+    [2**n for n in range(16, 66)]
+    # powers of 10
+    + [10**n for n in range(5, 20)]
+    # factorials
+    + [math.factorial(n) for n in range(9, 21)]
+    # a few primorial numbers https://en.wikipedia.org/wiki/Primorial
+    + [
+        510510,
+        6469693230,
+        304250263527210,
+        32589158477190044730,
+    ]
+)
+_constants_integers.extend(
+    [n - 1 for n in _constants_integers] + [n + 1 for n in _constants_integers]
+)
+_constants_integers.extend([-x for x in _constants_integers])
+
+# arbitrary cutoffs to keep our list bounded
+assert all(50_000 <= abs(n) <= 2**66 for n in _constants_integers)
+
 _constant_floats = (
     [
         0.5,
@@ -137,6 +194,8 @@ _constant_strings = {
     "if",
     "then",
     "else",
+    "__dict__",
+    "__proto__",  # javascript
     # strings which can be interpreted as a number
     "0",
     "1e100",
@@ -158,14 +217,14 @@ _constant_strings = {
     "Ⱥ",
     "Ⱦ",
     # ligatures
-    "æœÆŒﬀʤʨß"
+    "æœÆŒﬀʤʨß",
     # emoticons
     "(╯°□°）╯︵ ┻━┻)",
     # emojis
     "😍",
     "🇺🇸",
     # emoji modifiers
-    "🏻"  # U+1F3FB Light Skin Tone,
+    "🏻",  # U+1F3FB Light Skin Tone,
     "👍🏻",  # 👍 followed by U+1F3FB
     # RTL text
     "الكل في المجمو عة",
@@ -199,7 +258,7 @@ _constant_strings = {
 # we don't actually care what order the constants are sorted in, just that the
 # ordering is deterministic.
 GLOBAL_CONSTANTS = Constants(
-    integers=SortedSet(),
+    integers=SortedSet(_constants_integers),
     floats=SortedSet(_constant_floats, key=float_to_int),
     bytes=SortedSet(),
     strings=SortedSet(_constant_strings),
@@ -216,7 +275,7 @@ _local_constants = Constants(
 # modules are new without an expensive path.resolve() or is_local_module_file
 # cache lookup.
 _seen_modules: set[ModuleType] = set()
-_sys_modules_len: Optional[int] = None
+_sys_modules_len: int | None = None
 
 
 def _get_local_constants() -> Constants:
@@ -251,7 +310,10 @@ def _get_local_constants() -> Constants:
     # careful: store sys.modules length when we first check to avoid race conditions
     # with other threads loading a module before we set _sys_modules_len.
     if (sys_modules_len := len(sys.modules)) != _sys_modules_len:
-        new_modules = set(sys.modules.values()) - _seen_modules
+        # set(_seen_modules) shouldn't typically be required, but I have run into
+        # a "set changed size during iteration" error here when running
+        # test_provider_conformance_crosshair.
+        new_modules = set(sys.modules.values()) - set(_seen_modules)
         # Repeated SortedSet unions are expensive. Do the initial unions on a
         # set(), then do a one-time union with _local_constants after.
         new_constants = Constants()
@@ -273,47 +335,85 @@ def _get_local_constants() -> Constants:
     return _local_constants
 
 
+@contextmanager
+def with_register_backend(name, provider_cls):
+    try:
+        AVAILABLE_PROVIDERS[name] = provider_cls
+        yield
+    finally:
+        del AVAILABLE_PROVIDERS[name]
+
+
 class _BackendInfoMsg(TypedDict):
-    type: str
+    type: InfoObservationType
     title: str
-    content: Union[str, dict[str, Any]]
+    content: str | dict[str, Any]
+
+
+# TODO_DOCS: link to choice sequence explanation page
 
 
 class PrimitiveProvider(abc.ABC):
     """
-    A primitive provider is the implementation interface of a Hypothesis backend
-    (see :ref:`alternative-backends`).
+    |PrimitiveProvider| is the implementation interface of a
+    :ref:`Hypothesis backend <alternative-backends>`.
+
+    A |PrimitiveProvider| is required to implement the following five
+    ``draw_*`` methods:
+
+    * |PrimitiveProvider.draw_integer|
+    * |PrimitiveProvider.draw_boolean|
+    * |PrimitiveProvider.draw_float|
+    * |PrimitiveProvider.draw_string|
+    * |PrimitiveProvider.draw_bytes|
+
+    Each strategy in Hypothesis generates values by drawing a series of choices
+    from these five methods. By overriding them, a |PrimitiveProvider| can control
+    the distribution of inputs generated by Hypothesis.
+
+    For example, :pypi:`hypothesis-crosshair` implements a |PrimitiveProvider|
+    which uses an SMT solver to generate inputs that uncover new branches.
+
+    Once you implement a |PrimitiveProvider|, you can make it available for use
+    through |AVAILABLE_PROVIDERS|.
     """
 
-    # How long a provider instance is used for. One of test_function or
-    # test_case. Defaults to test_function.
-    #
-    # If test_function, a single provider instance will be instantiated and used
-    # for the entirety of each test function. I.e., roughly one provider per
-    # @given annotation. This can be useful if you need to track state over many
-    # executions to a test function.
-    #
-    # This lifetime will cause None to be passed for the ConjectureData object
-    # in PrimitiveProvider.__init__, because that object is instantiated per
-    # test case.
-    #
-    # If test_case, a new provider instance will be instantiated and used each
-    # time hypothesis tries to generate a new input to the test function. This
-    # lifetime can access the passed ConjectureData object.
-    #
-    # Non-hypothesis providers probably want to set a lifetime of test_function.
-    lifetime: _Lifetime = "test_function"
+    #: The lifetime of a |PrimitiveProvider| instance. Either ``test_function``
+    #: or ``test_case``.
+    #:
+    #: If ``test_function`` (the default), a single provider instance will be
+    #: instantiated and used for the entirety of each test function (i.e., roughly
+    #: one provider per |@given| annotation). This can be useful for tracking state
+    #: over the entirety of a test function.
+    #:
+    #: If ``test_case``, a new provider instance will be instantiated and used for
+    #: each input Hypothesis generates.
+    #:
+    #: The ``conjecturedata`` argument to ``PrimitiveProvider.__init__`` will
+    #: be ``None`` for a lifetime of ``test_function``, and an instance of
+    #: ``ConjectureData`` for a lifetime of ``test_case``.
+    #:
+    #: Third-party providers likely want to set a lifetime of ``test_function``.
+    lifetime: ClassVar[LifetimeT] = "test_function"
 
-    # Solver-based backends such as hypothesis-crosshair use symbolic values
-    # which record operations performed on them in order to discover new paths.
-    # If avoid_realization is set to True, hypothesis will avoid interacting with
-    # symbolic choices returned by the provider in any way that would force the
-    # solver to narrow the range of possible values for that symbolic.
-    #
-    # Setting this to True disables some hypothesis features, such as
-    # DataTree-based deduplication, and some internal optimizations, such as
-    # caching constraints. Only enable this if it is necessary for your backend.
-    avoid_realization = False
+    #: Solver-based backends such as ``hypothesis-crosshair`` use symbolic values
+    #: which record operations performed on them in order to discover new paths.
+    #: If ``avoid_realization`` is set to ``True``, hypothesis will avoid interacting
+    #: with symbolic choices returned by the provider in any way that would force
+    #: the solver to narrow the range of possible values for that symbolic.
+    #:
+    #: Setting this to ``True`` disables some hypothesis features and optimizations.
+    #: Only set this to ``True`` if it is necessary for your backend.
+    avoid_realization: ClassVar[bool] = False
+
+    #: If ``True``, |PrimitiveProvider.on_observation| will be added as a
+    #: callback via |add_observability_callback|, enabling observability during
+    # the lifetime of this provider. If ``False``, |PrimitiveProvider.on_observation|
+    #: will never be called by Hypothesis.
+    #:
+    #: The opt-in behavior of observability is because enabling observability
+    #: might increase runtime or memory usage.
+    add_observability_callback: ClassVar[bool] = False
 
     def __init__(self, conjecturedata: Optional["ConjectureData"], /) -> None:
         self._cd = conjecturedata
@@ -323,18 +423,49 @@ class PrimitiveProvider(abc.ABC):
         self,
         p: float = 0.5,
     ) -> bool:
+        """
+        Draw a boolean choice.
+
+        Parameters
+        ----------
+        p: float
+            The probability of returning ``True``. Between 0 and 1 inclusive.
+
+            Except for ``0`` and ``1``, the value of ``p`` is a hint provided by
+            Hypothesis, and may be ignored by the backend.
+
+            If ``0``, the provider must return ``False``. If ``1``, the provider
+            must return ``True``.
+        """
         raise NotImplementedError
 
     @abc.abstractmethod
     def draw_integer(
         self,
-        min_value: Optional[int] = None,
-        max_value: Optional[int] = None,
+        min_value: int | None = None,
+        max_value: int | None = None,
         *,
-        # weights are for choosing an element index from a bounded range
-        weights: Optional[dict[int, float]] = None,
+        weights: dict[int, float] | None = None,
         shrink_towards: int = 0,
     ) -> int:
+        """
+        Draw an integer choice.
+
+        Parameters
+        ----------
+        min_value : int | None
+            (Inclusive) lower bound on the integer value. If ``None``, there is
+            no lower bound.
+        max_value : int | None
+            (Inclusive) upper bound on the integer value. If ``None``, there is
+            no upper bound.
+        weights: dict[int, float] | None
+            Maps keys in the range [``min_value``, ``max_value``] to the probability
+            of returning that key.
+        shrink_towards: int
+            The integer to shrink towards. This is not used during generation and
+            can be ignored by backends.
+        """
         raise NotImplementedError
 
     @abc.abstractmethod
@@ -346,6 +477,21 @@ class PrimitiveProvider(abc.ABC):
         allow_nan: bool = True,
         smallest_nonzero_magnitude: float,
     ) -> float:
+        """
+        Draw a float choice.
+
+        Parameters
+        ----------
+        min_value : float
+            (Inclusive) lower bound on the float value.
+        max_value : float
+            (Inclusive) upper bound on the float value.
+        allow_nan : bool
+            If ``False``, it is invalid to return ``math.nan``.
+        smallest_nonzero_magnitude : float
+            The smallest allowed nonzero magnitude. ``draw_float`` should not
+            return a float ``f`` if ``abs(f) < smallest_nonzero_magnitude``.
+        """
         raise NotImplementedError
 
     @abc.abstractmethod
@@ -356,6 +502,18 @@ class PrimitiveProvider(abc.ABC):
         min_size: int = 0,
         max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
     ) -> str:
+        """
+        Draw a string choice.
+
+        Parameters
+        ----------
+        intervals : IntervalSet
+            The set of codepoints to sample from.
+        min_size : int
+            (Inclusive) lower bound on the string length.
+        max_size : int
+            (Inclusive) upper bound on the string length.
+        """
         raise NotImplementedError
 
     @abc.abstractmethod
@@ -364,9 +522,19 @@ class PrimitiveProvider(abc.ABC):
         min_size: int = 0,
         max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
     ) -> bytes:
+        """
+        Draw a bytes choice.
+
+        Parameters
+        ----------
+        min_size : int
+            (Inclusive) lower bound on the bytes length.
+        max_size : int
+            (Inclusive) upper bound on the bytes length.
+        """
         raise NotImplementedError
 
-    def per_test_case_context_manager(self):
+    def per_test_case_context_manager(self) -> AbstractContextManager:
         """
         Returns a context manager which will be entered each time Hypothesis
         starts generating and executing one test case, and exited when that test
@@ -377,7 +545,7 @@ class PrimitiveProvider(abc.ABC):
         generating strategy values for each test case. This is just before any
         :ref:`custom executor <custom-function-execution>` is called.
 
-        Even if not returning a custom context manager, PrimitiveProvider
+        Even if not returning a custom context manager, |PrimitiveProvider|
         subclasses are welcome to override this method to know when Hypothesis
         starts and ends the execution of a single test case.
         """
@@ -386,19 +554,19 @@ class PrimitiveProvider(abc.ABC):
     def realize(self, value: T, *, for_failure: bool = False) -> T:
         """
         Called whenever hypothesis requires a concrete (non-symbolic) value from
-        a potentially symbolic value. Hypothesis will not check that `value` is
-        symbolic before calling `realize`, so you should handle the case where
-        `value` is non-symbolic.
+        a potentially symbolic value. Hypothesis will not check that ``value`` is
+        symbolic before calling ``realize``, so you should handle the case where
+        ``value`` is non-symbolic.
 
         The returned value should be non-symbolic.  If you cannot provide a value,
-        raise hypothesis.errors.BackendCannotProceed("discard_test_case").
+        raise |BackendCannotProceed| with a value of ``"discard_test_case"``.
 
-        If for_failure is True, the value is associated with a failing example.
+        If ``for_failure`` is ``True``, the value is associated with a failing example.
         In this case, the backend should spend substantially more effort when
         attempting to realize the value, since it is important to avoid discarding
-        failing  examples. Backends may still raise BackendCannotProceed when
-        for_failure is True, if realization is truly impossible or if realization
-        takes significantly longer than expected (say, 5 minutes).
+        failing examples. Backends may still raise |BackendCannotProceed| when
+        ``for_failure`` is ``True``, if realization is truly impossible or if
+        realization takes significantly longer than expected (say, 5 minutes).
         """
         return value
 
@@ -408,50 +576,136 @@ class PrimitiveProvider(abc.ABC):
         may wish to enqueue to replay under its own instrumentation when we next
         ask to generate a test case, rather than generating one from scratch.
 
-        This is used to e.g. warm-start hypothesis-crosshair with a corpus of
-        high-code-coverage inputs discovered by HypoFuzz.
+        This is used to e.g. warm-start :pypi:`hypothesis-crosshair` with a corpus
+        of high-code-coverage inputs discovered by
+        `HypoFuzz <https://hypofuzz.com/>`_.
         """
         return None
 
     def observe_test_case(self) -> dict[str, Any]:
-        """Called at the end of the test case when observability mode is active.
+        """Called at the end of the test case when :ref:`observability
+        <observability>` is enabled.
 
         The return value should be a non-symbolic json-encodable dictionary,
-        and will be included as `observation["metadata"]["backend"]`.
+        and will be included in observations as ``observation["metadata"]["backend"]``.
         """
         return {}
 
     def observe_information_messages(
-        self, *, lifetime: _Lifetime
+        self, *, lifetime: LifetimeT
     ) -> Iterable[_BackendInfoMsg]:
         """Called at the end of each test case and again at end of the test function.
 
-        Return an iterable of `{type: info/alert/error, title: str, content: str|dict}`
-        dictionaries to be delivered as individual information messages.
-        (Hypothesis adds the `run_start` timestamp and `property` name for you.)
+        Return an iterable of ``{type: info/alert/error, title: str, content: str | dict}``
+        dictionaries to be delivered as individual information messages. Hypothesis
+        adds the ``run_start`` timestamp and ``property`` name for you.
         """
         assert lifetime in ("test_case", "test_function")
         yield from []
 
-    def span_start(self, label: int, /) -> None:  # noqa: B027  # non-abstract noop
-        """Marks the beginning of a semantically meaningful span.
+    def on_observation(self, observation: TestCaseObservation) -> None:  # noqa: B027
+        """
+        Called at the end of each test case which uses this provider, with the same
+        ``observation["type"] == "test_case"`` observation that is passed to
+        other callbacks added via |add_observability_callback|. This method is not
+        called with ``observation["type"] in {"info", "alert", "error"}``
+        observations.
 
-        Providers can optionally track this data to learn which sub-sequences
-        of draws correspond to a higher-level object, recovering the parse tree.
-        `label` is an opaque integer, which will be shared by all spans drawn
-        from a particular strategy.
+        .. important::
 
-        This method is called from ConjectureData.start_span().
+            For |PrimitiveProvider.on_observation| to be called by Hypothesis,
+            |PrimitiveProvider.add_observability_callback| must be set to ``True``.
+
+            |PrimitiveProvider.on_observation| is explicitly opt-in, as enabling
+            observability might increase runtime or memory usage.
+
+        Calls to this method are guaranteed to alternate with calls to
+        |PrimitiveProvider.per_test_case_context_manager|. For example:
+
+        .. code-block:: python
+
+            # test function starts
+            per_test_case_context_manager()
+            on_observation()
+            per_test_case_context_manager()
+            on_observation()
+            ...
+            # test function ends
+
+        Note that |PrimitiveProvider.on_observation| will not be called for test
+        cases which did not use this provider during generation, for example
+        during |Phase.reuse| or |Phase.shrink|, or because Hypothesis switched
+        to the standard Hypothesis backend after this backend raised too many
+        |BackendCannotProceed| exceptions.
         """
 
-    def span_end(self, discard: bool, /) -> None:  # noqa: B027, FBT001
-        """Marks the end of a semantically meaningful span.
+    def span_start(self, label: int, /) -> None:  # noqa: B027  # non-abstract noop
+        """Marks the beginning of a semantically meaningful span of choices.
 
-        `discard` is True when the draw was filtered out or otherwise marked as
-        unlikely to contribute to the input data as seen by the user's test.
+        Spans are a depth-first tree structure. A span is opened by a call to
+        |PrimitiveProvider.span_start|, and a call to |PrimitiveProvider.span_end|
+        closes the most recently opened span. So the following sequence of calls:
+
+        .. code-block:: python
+
+            span_start(label=1)
+            n1 = draw_integer()
+            span_start(label=2)
+            b1 = draw_boolean()
+            n2 = draw_integer()
+            span_end()
+            f1 = draw_float()
+            span_end()
+
+        produces the following two spans of choices:
+
+        .. code-block::
+
+            1: [n1, b1, n2, f1]
+            2: [b1, n2]
+
+        Hypothesis uses spans to denote "semantically meaningful" sequences of
+        choices. For instance, Hypothesis opens a span for the sequence of choices
+        made while drawing from each strategy. Not every span corresponds to a
+        strategy; the generation of e.g. each element in |st.lists| is also marked
+        with a span, among others.
+
+        ``label`` is an opaque integer, which has no defined semantics.
+        The only guarantee made by Hypothesis is that all spans with the same
+        "meaning" will share the same ``label``. So all spans from the same
+        strategy will share the same label, as will e.g. the spans for |st.lists|
+        elements.
+
+        Providers can track calls to |PrimitiveProvider.span_start| and
+        |PrimitiveProvider.span_end| to learn something about the semantics of
+        the test's choice sequence. For instance, a provider could track the depth
+        of the span tree, or the number of unique labels, which says something about
+        the complexity of the choices being generated. Or a provider could track
+        the span tree across test cases in order to determine what strategies are
+        being used in what contexts.
+
+        It is possible for Hypothesis to start and immediately stop a span,
+        without calling a ``draw_*`` method in between. These spans contain zero
+        choices.
+
+        Hypothesis will always balance the number of calls to
+        |PrimitiveProvider.span_start| and |PrimitiveProvider.span_end|. A call
+        to |PrimitiveProvider.span_start| will always be followed by a call to
+        |PrimitiveProvider.span_end| before the end of the test case.
+
+        |PrimitiveProvider.span_start| is called from ``ConjectureData.start_span()``
+        internally.
+        """
+
+    def span_end(self, discard: bool, /) -> None:  # noqa: B027
+        """Marks the end of a semantically meaningful span of choices.
+
+        ``discard`` is ``True`` when the draw was filtered out or otherwise marked
+        as unlikely to contribute to the input data as seen by the user's test.
         Note however that side effects can make this determination unsound.
 
-        This method is called from ConjectureData.stop_span().
+        |PrimitiveProvider.span_end| is called from ``ConjectureData.stop_span()``
+        internally.
         """
 
 
@@ -475,14 +729,15 @@ class HypothesisProvider(PrimitiveProvider):
         p: float = 0.05,
     ) -> Optional["ConstantT"]:
         assert self._random is not None
-        assert self._local_constants is not None
         assert choice_type != "boolean"
-
         # check whether we even want a constant before spending time computing
         # and caching the allowed constants.
         if self._random.random() > p:
             return None
 
+        # note: this property access results in computation being done
+        assert self._local_constants is not None
+
         key = (choice_type, choice_constraints_key(choice_type, constraints))
         if key not in CONSTANTS_CACHE:
             CONSTANTS_CACHE[key] = (
@@ -530,10 +785,10 @@ class HypothesisProvider(PrimitiveProvider):
 
     def draw_integer(
         self,
-        min_value: Optional[int] = None,
-        max_value: Optional[int] = None,
+        min_value: int | None = None,
+        max_value: int | None = None,
         *,
-        weights: Optional[dict[int, float]] = None,
+        weights: dict[int, float] | None = None,
         shrink_towards: int = 0,
     ) -> int:
         assert self._cd is not None
@@ -847,10 +1102,10 @@ class BytestringProvider(PrimitiveProvider):
 
     def draw_integer(
         self,
-        min_value: Optional[int] = None,
-        max_value: Optional[int] = None,
+        min_value: int | None = None,
+        max_value: int | None = None,
         *,
-        weights: Optional[dict[int, float]] = None,
+        weights: dict[int, float] | None = None,
         shrink_towards: int = 0,
     ) -> int:
         assert self._cd is not None
